#!/usr/bin/env python
#------------------------------------------------------------------------------
# Author:   Pierre Schnizer <Pierre.Schnizer@cern.ch>
# Date:     05. 11. 2002
# Purpose:  To illustrate the usage and capability of f2py
# Version:  Not even alpha. This is illustration code, so do not use it in
#           production environement
# License:  LGPL
#------------------------------------------------------------------------------
import Numeric
import MLab
import _lsodar

class _LastJob:
    jobid = -2
    joblist = {}
            
    def GetLastJobId(self):
        return self.jobid

    def GetJob(self, id):
        return self.joblist[id]

    def Register(self, instance):
        tmp = id(instance)
        self.joblist[tmp] = instance
        return tmp

    def DeRegister(self, instance):
        tmp = id(instance)
        del self.joblist[tmp]
        
    def SetActiveJobId(self, id):
        self.jobid = id
    
_lastjob = _LastJob()

_msgs = {2: "Integration successful.",
         3: "Found Border",
         -1: "Excess work done on this call (perhaps wrong Dfun type).",
         -2: "Excess accuracy requested (tolerances too small).",
         -3: "Illegal input detected (internal error).",
         -4: "Repeated error test failures (internal error).",
         -5: "Repeated convergence failures (perhaps bad Jacobian or"+\
         " tolerances).",
         -6: "Error weight became zero during problem.",
         -7: "Internal workspace insufficient to finish (internal error)."
         }


#--------------------- Class Integrate ----------------------------------------
class Integrate:
    """
    Wraps lsodar. Allows multiple instances.

    LSODAR from the ODEPACK package provides a flexible ODE solver. It solves a
    system of first order differential equations. Further it searches for the
    roots in a function specified as border.

    For each problem you should set up a seperate object. Between calls the
    internal state is preserved.

    Currently jacobian matrices can not be supplied.
    """
    def __del__(self):
        self._lastjob.Deregister(self)
        
    def __init__(self, integrand, y0, t0, **dic):
        """
        Input : integrand ... the function system describing the problem
                y0        ... the start vector for the dependent variables
                t0        ... the start value for the independen variable

        optional: border  ... a funtion returning an array. When one of its
                              values is zero,  the evaluation will terminate.
        """
        self._lastjob = _lastjob
        self.jobid = self._lastjob.Register(self)
        assert(id(self) == id(self._lastjob.GetJob(self.jobid)))
        
        border = None
        args = None
        borderargs = None
        y0 = y0
        t0 = t0
        if dic.has_key('border'):
            border = dic['border']
        if dic.has_key('args'):
            args = tuple(dic['args'])
        if dic.has_key('borderargs'):
            borderargs = tuple(dic['borderargs'])    
            
        self.args = args
        self.borderargs = borderargs
        tmp = (t0, y0)
        if args != None:
            tmp += args
        dimension = len(Numeric.array(apply(integrand, tmp)))
        if border != None:
            tmp = (t0, y0)
            if borderargs != None:
                tmp += borderargs
            self.borderdimension = len(Numeric.array(apply(border, tmp)))
            self.border = border
        else:
            self.borderdimension = 0
            self.border = lambda x: x
        self.integrand = integrand    
        self.rsav  = Numeric.zeros((245), Numeric.Float)
        self.isav  = Numeric.zeros((59), Numeric.Int)
        rlength = 22 + dimension * MLab.max((16, dimension+9))
        rlength += 3 * self.borderdimension
        ilength = 20 + dimension
        self.rwork = Numeric.zeros((rlength), Numeric.Float)
        self.iwork = Numeric.zeros((ilength), Numeric.Int)

        self.dimension = dimension
        self.state = 1
        self.itol = 1
        self.rtol = Numeric.array(1e-6)
        self.atol = Numeric.array(1e-6)
        self.iopt = 0
        self.y = y0
        self.t = t0

        # Enable Common Block handling of lsodar.f
        self.needpostcleanup = -1

        #print "id of _lastjob : ", id(self._lastjob),
        #print "Printed from init :",id(self)
        
    def _SetPrecision(self, which, precision):
        tmp = Numeric.array(precision)
        length = len(tmp)
        assert(length == 1 or length == self.dimension)
        if which == 'absolute':
            self.atol = tmp
        if which == 'relative':
            self.rtol = tmp
        al = len(self.atol)
        rl = len(self.rtol)    
        if length > 1 or al > 1 or rl > 1:
            self.itol = 2
        else:
            self.itol = 1
        if self.itol == 2:    
            #Multidimensional ...            
            if al == 1:
                self.atol = self.atol * Numeric.ones(self.dimension)
            if il == 1:
                self.itol = self.itol * Numeric.ones(self.dimension)    
                
        
    def SetAbsolutePrecision(self, precision):
        self._SetPrecision('absolute', precision)

        
    def SetRelativePrecision(self, precision):
        self._SetPrecision('relative', precision)




    def Reinit(self, y0, t0):
        self.state = 1
        assert(len(y0) == self.dimension)
        self.y = y0
        self.t = t0
        
    def __call__(self, t, **dic):
        """
        The calculation routine.

        Input:  t    ...  values of the independent variable.
                          The dependent variable will be caluclate
                          for all values of t[1:]
                          So the length of t must be 2 at minimum
                          
        Keywords:  task: an index specifying the task to be performed.  
                          task has the following values and meanings.
                          1  means normal computation of output values of y(t)
                             at t = tout (by overshooting and interpolating).
                          2  means take one step only and return.
                          3  means stop at the first internal mesh point at or
                             beyond t = tout and return.
                          If you specifiy a critical point in evaluation only
                          value 1 and 2 are allowed.
                          
                   tcrit: Defines a critical point,  where the problem should
                          not be evaluted. tcrit may be equal to or beyond the
                          last value of t, but not behind it in the direction
                          of integration.  this option is useful if the problem
                          has a singularity at or beyond t = tcrit.


        Output: yout, t, jroot
                yout   ...   the values corresponding to t
                t      ...   passed to see what value t had if a border was
                             found
                jroot  ...   indicates which of the border conditions
                             has been fulfilled. The last values
                             of t an
                             
        !!!!!!!!!!!!!!!!!!!!
        !!!   Warning   !!!!  
        !!!!!!!!!!!!!!!!!!!!

        Save's the system state internally. So use one object for each problem!
        """
        task = 1
        if dic.has_key('task'):
            task = dic['task']
        if dic.has_key('tcrit'):
            if task !=  1 and task != 2 :
                des = "Wrong Value of task when calculating with a critcal "+\
                "time. Task was " + str(task) + " but only 1 or 2 are allowed!"
                raise ValueError, des
            task = task + 3
            self.rwork[1 - 1] = dic['tcrit']
        else:
            if task !=  1 and task != 2 and task != 3:
                des = "Wrong Value of task when calculating with a critcal"+\
                      " time. Task was " + str(task) + " but only 1, 2  or"+\
                      " 3 are allowed!"
                raise ValueError, des
            
        iopt = self.iopt

        # Callback to jacobi not supported yet!
        jacobi = lambda x:x
        jt = 2

        # Handle additional arguments to the functions
        integrand = self.integrand
        border    = self.border
        if self.args != None:
            def integrand(t,y, tmp, myint=self.integrand, myargs=self.args):
                tmp1 =  apply(myint, (t,y) + myargs)
                return tmp1
            
        if self.borderargs != None:
            def border(t,y, tmp, myb = self.border, myargs=self.borderargs):
                return apply(myb, (t,y) + myargs)

        # Handle a vector input of y
        result = None
        if type(t) == Numeric.arraytype:
            result = Numeric.zeros((len(self.y),
                                    t.shape[0]), Numeric.Float)
            
        # Call the FORTRAN routine
        self.__prae() # Write Common Blocks
        if result == None:
            # Only one datum for t
            y = self.y
            t0 = self.t
            tmp = _lsodar.lsodar(integrand, y, t0, t,
                                 # Type of tolerances and tolerances
                                 self.itol, self.rtol, self.atol,
                                 # Task and status of calculation
                                 task, self.state,
                                 # Optional paramters and work arrays
                                 iopt, self.rwork, self.iwork,
                                 # Jacobi matrix, external not supported yet
                                 jacobi, jt,
                                 # Border conditions
                                 border, self.borderdimension)
            yout, t, state, jroot = tmp
        else:
            # handle a vector
            y = self.y
            # Do vector operations ...
            for i in Numeric.arange(t.shape[0]):
                if i == 0:
                    t0 = self.t
                else:
                    t0 = t[i-1]                    
                tmp=_lsodar.lsodar(integrand, y, t0, t[i],
                                   # Type of tolerances and tolerances
                                   self.itol, self.rtol, self.atol,
                                   # Task and status of calculation
                                   task, self.state,
                                   # Optional paramters and work arrays
                                   iopt, self.rwork, self.iwork,
                                   # Jacobi matrix, external not supported yet
                                   jacobi, jt,
                                   # Border conditions
                                   border, self.borderdimension)
                y, tend, state, jroot = tmp
                result[:,i] = y                
                if state < 1:
                    # Error -> raise Error
                    break
                if state == 3:
                    break
                 
                # Important: Must be inside for loop
                self.state = state
                # End Of for loop -- to big loop ..
            if state == 2 or state == 1:    
                yout = result
            elif state == 3:
                # Save the time when a border was encountered
                t[i] = tend
                t = t[:i+1]
                yout = result[:,:i+1]
            # End of if  -- toooo big conditional
        
        
        self.__post() # Save Common Block Status


        # Handle errors from the routine
        if state < 1:
            raise ValueError, _msgs[state]
        if result == None:
            self.y = yout
            self.t = t
        else:
            self.y = yout[:,-1]
            self.t = t[-1]
            
        if state == 3:
            # Border condition ...
            self.state = 2
            return yout, jroot, t
        
        self.state = state
        return yout, None, None
        


    def Derivative(self, t, order):
        """
        Calculates the order derivative at the value t.

        !!!!!!!!!!!!!!!!!!!!
        !!!   Warning   !!!!  
        !!!!!!!!!!!!!!!!!!!!
        Can only be used after the calculation!
        """
        assert (self.state == 2 or self.state == 3)
        self.__prae()
        tmp = _lsodar.intdy(t, order, self.rwork, self.dimension)
        self.__post()
        return tmp

    def _SaveCommonBlockData(self):
        # Load common's information to arrays
        _lsodar.srcar(self.rsav, self.isav, 1)

    def _StoreCommonBlockData(self):
        # Restore common's from arrays
        _lsodar.srcar(self.rsav, self.isav, 2)
        
    def __prae(self):
        #print "id of _lastjob : ", id(self._lastjob),
        #print "Printed from prae : ",id(self)
        tmp = self._lastjob.GetLastJobId()

        # Jobids lower than 0 mean nothing needs to be done
        if  tmp != self.jobid and tmp > 0:
            # Tell the other job to save its data
            self._lastjob.GetJob(tmp)._SaveCommonBlockData()
            self._StoreCommonBlockData()
            self.needpostcleanup = tmp
        else:
            self.needpostcleanup = 0
        #print "self.needpostcleanup = ", self.needpostcleanup
        
        self._lastjob.SetActiveJobId(self.jobid)
        
    def __post(self):
        #print "id of _lastjob : ", id(self._lastjob),
        #print "Printed from post : ",id(self)
        if self.needpostcleanup > 0:
            self._SaveCommonBlockData()
            tmp = self._lastjob.GetJob(self.needpostcleanup)
            tmp._StoreCommonBlockData()
            self._lastjob.SetActiveJobId(self.needpostcleanup)
            self.needpostcleanup = -1
        elif self.needpostcleanup == 0:
            pass
        else:
            raise TypeError, "self.needpostcleanup in instance " + `id(self)`+\
                  " should be 0 or 1 \n but was found to: " +\
                  `self.needpostcleanup`
        

